from libcst import CSTTransformer, ClassDef, FunctionDef, Module, SimpleStatementLine, Import, ImportFrom


class MergingTransformer(CSTTransformer):
    def __init__(self, new_cst):
        super().__init__()
        self.new_cst = new_cst
        self.stack = []
        self._push(new_cst.body)

    def _push(self, new_members):
        self.stack.append((new_members, {}))

    def _pop(self):
        return self.stack.pop()

    def _set(self, key, value):
        self.stack[-1][1][key] = value

    def visit_ClassDef(self, node: ClassDef):
        self._set(node.name.value, node)
        for part in self.stack[-1][0]:
            if hasattr(part, 'name'):
                if part.name.value == node.name.value:
                    if hasattr(part, 'body'):
                        self._push(part.body.body)
                        break
        else:
            self._push([])

    def visit_FunctionDef(self, node: FunctionDef):
        # do not visit the children of the function as we do not modify
        # an already written function code
        self._set(node.name.value, node)
        return False

    def leave_ClassDef(
            self, original_node: ClassDef, updated_node: ClassDef
    ):
        parts_to_add = self.leave_class_module()
        self._pop()

        if parts_to_add:
            # and merge them with the original content
            return updated_node.with_changes(
                body=updated_node.body.with_changes(
                    body=[
                        *updated_node.body.body,
                        *parts_to_add
                    ]
                )
            )
        return updated_node

    def leave_Module(self, original_node: Module, updated_node: Module):
        parts_to_add = self.leave_class_module()

        new_imports = self._extract_imports(self.stack[0][0])
        existing_imports = self._extract_imports(original_node.body)

        # extract imports that are not yet present
        extra_imports = []
        for ni in new_imports:
            for ei in existing_imports:
                if ei.deep_equals(ni):
                    break
            else:
                extra_imports.append(ni)

        self._pop()

        # and merge
        return updated_node.with_changes(
            body=[
                *extra_imports,
                *updated_node.body,
                *parts_to_add
            ]
        )

    def _extract_imports(self, lines):
        imports = []
        for l in lines:
            if isinstance(l, SimpleStatementLine):
                for ml in l.children:
                    if isinstance(ml, (Import, ImportFrom)):
                        imports.append(l)
                        break
        return imports

    def leave_class_module(self):
        new_parts = self.stack[-1][0]
        existing_parts = self.stack[-1][1]

        # look for the parts in the new parts that are not in the existing parts
        parts_to_add = []
        for part in new_parts:
            if hasattr(part, 'name'):
                if part.name.value not in existing_parts:
                    parts_to_add.append(part)
        return parts_to_add



from collections import defaultdict

from oarepo_model_builder.builders import process
from oarepo_model_builder.stack import ModelBuilderStack
from .invenio_base import InvenioBaseClassPythonBuilder
from ..outputs.json_stack import JSONStack
from ..utils.deepmerge import deepmerge
from ..utils.schema import is_schema_element


class InvenioRecordSchemaBuilder(InvenioBaseClassPythonBuilder):
    TYPE = 'invenio_record_schema'
    class_config = 'record-schema-class'
    template = 'record-schema'

    def begin(self, schema, settings):
        super().begin(schema, settings)
        self.stack = JSONStack()
        self.imports = defaultdict(set)  # import => aliases
        self.imports['marshmallow'].add('ma')
        self.imports['marshmallow.fields'].add('ma_fields')
        self.imports['marshmallow.validate'].add('ma_valid')

    def finish(self):
        # TODO: generate subschemas as well
        # TODO: generate arrays as well
        # TODO: handle required
        super().finish(
            fields=self.stack.value.get('properties', {}),
            imports=self.imports,

        )

    @process('/model/**', condition=lambda current: is_schema_element(current.stack))
    def enter_model_element(self, stack: ModelBuilderStack):
        self.model_element_enter(stack)

        # process children
        yield

        data = stack.top.data
        if isinstance(data, dict):
            set_definition = False
            if 'oarepo:marshmallow' in data:
                self.stack.push('oarepo:marshmallow', data['oarepo:marshmallow'])
                self.stack.pop()
                set_definition = True
            elif stack.stack[-2].key == 'properties':
                self.stack.push(
                    'oarepo:marshmallow', {})

                self.stack.pop()
                set_definition = True
            if set_definition:
                self.set_marshmallow_definition(self.stack.real_top)

        self.model_element_leave(stack)

    def set_marshmallow_definition(self, data):
        marshmallow = data['oarepo:marshmallow']

        data_type = data.get('type', 'object')
        generator = self.schema.settings.python.marshmallow.mapping.get(data_type, None)
        if isinstance(generator, dict):
            # a value, not a callable => merge it
            marshmallow = deepmerge(generator, marshmallow)
            data['marshmallow'] = marshmallow
            generator = None

        # add imports if required
        for imp in marshmallow.get('imports', []):
            self.imports[imp['import']].add(imp['alias'])

        if 'field' in marshmallow:
            # do not modify the field
            return
        if 'class' in marshmallow:
            # generate instance of the class, filling the options and validators
            marshmallow['field'] = create_field(marshmallow['class'], options=(), validators=(), data=data)
            return

        # if no generator from settings, get the default one
        if not generator:
            generator = default_marshmallow_generators.get(data_type, None)
            if not generator:
                raise Exception(f'Do not have marshmallow field generator for {data_type}. '
                                f'Define it either in invenio_record_schema.py or in your own config')

        # and generate the field
        marshmallow['field'] = generator(data, self.schema, self.imports)

    def model_element_enter(self, stack: ModelBuilderStack):
        top = stack.top
        match stack.top_type:
            case stack.PRIMITIVE:
                self.stack.push(top.key, top.data)
            case stack.LIST:
                self.stack.push(top.key, [])
            case stack.DICT:
                self.stack.push(top.key, {})

    def model_element_leave(self, stack: ModelBuilderStack):
        self.stack.pop()


def create_field(field_type, options=(), validators=(), data=None):
    opts = [*options, *data['oarepo:marshmallow'].get('options', [])]
    validators = [*validators, *data['oarepo:marshmallow'].get('validators', [])]
    nested = data['oarepo:marshmallow'].get('nested', False)
    if validators:
        opts.append(f'validate=[{",".join(validators)}]')
    ret = f'{field_type}({", ".join(opts)})'
    if nested:
        ret = f'ma_fields.Nested({ret})'
    return ret


def marshmallow_string_generator(data, schema, imports):
    validators = []
    min_length = data.get('minLength', None)
    max_length = data.get('maxLength', None)
    if min_length is not None or max_length is not None:
        validators.append(f'ma_valid.Length(min={min_length}, max={max_length})')
    return create_field('ma_fields.String', [], validators, data)


# TODO: rest of supported schema types


default_marshmallow_generators = {
    'string': marshmallow_string_generator
}



from .invenio_base import InvenioBaseClassPythonBuilder


class InvenioBlueprintBuilder(InvenioBaseClassPythonBuilder):
    TYPE = 'invenio_blueprint'
    class_config = 'register-blueprint-function'
    template = 'blueprint'



import copy
import importlib
from pathlib import Path
from typing import List, Dict, Type

from .builders import OutputBuilder, ModelBuilderStack, ReplaceElement
from .outputs import OutputBase
from .property_preprocessors import PropertyPreprocessor
from .schema import ModelSchema
from .model_preprocessors import ModelPreprocessor


class ModelBuilder:
    """
    Processes a model file and generates/updates sources for the model
    """

    model_preprocessor_classes: List[type(ModelPreprocessor)]
    """
    Model preprocessor classes that are called after schema is loaded and before it is processed
    """

    output_classes: List[type(OutputBase)]
    """
    Mapping between output type and its handler class
    """

    filtered_output_classes: Dict[str, type(OutputBase)]
    """
    Filtered output classes by settings.plugins.disabled|enabled
    """

    output_builder_classes: List[type(OutputBuilder)]
    """
    A list of extension classes to be used in build. 
    """

    property_preprocessor_classes: List[type(PropertyPreprocessor)]
    """
    Processor classes (called before and after file builder is called)
    """

    output_builders: List[OutputBuilder]
    """
    A list of output_builders. Each extension is responsible for generating one or more files
    """

    outputs: Dict[Path, OutputBase]
    """
    Mapping between concrete output (file path relative to output dir) and instance of builder class
    """

    property_preprocessors: List[PropertyPreprocessor]
    """
    Current instances of processor classes.
    """

    def __init__(
            self,
            outputs: List[type(OutputBase)] = (),
            output_builders: List[type(OutputBuilder)] = (),
            property_preprocessors: List[type(PropertyPreprocessor)] = (),
            model_preprocessors: List[type(ModelPreprocessor)] = (),
            open=open
    ):
        """
        Initializes the builder

        :param output_builders:          A list of extension classes to use in builds
        :param outputs:     List of file builder classes that generate files
        :param property_preprocessors: List of output type processor classes
        """
        self.output_builder_classes = [*output_builders]
        for o in outputs:
            assert o.TYPE, f'output_type not set up on class {o}'
        self.output_classes = [*outputs]
        self.property_preprocessor_classes = [*property_preprocessors]
        self.model_preprocessor_classes = [*model_preprocessors]
        self.filtered_output_classes = {o.TYPE: o for o in self.output_classes}
        self.open = open

    def get_output(self, output_type: str, path: str | Path):
        """
        Given a path, instantiate file builder on the path with the given output type
        and return it. If the builder on the path has already been requested, return
        the same instance of the builder.

        :param output_type: @see FileBuilder.output_type
        :param path: relative path to output_dir, set in build()
        :return:    instance of FileBuilder for the path
        """
        if not isinstance(path, Path):
            path = Path(path)
        path = self.output_dir.joinpath(path)

        output = self.outputs.get(path, None)
        if output:
            assert output_type == self.outputs[path].TYPE
        else:
            output = self.filtered_output_classes[output_type](self, path)
            output.begin()
            self.outputs[path] = output
        return output

    # main entry point
    def build(self, schema: ModelSchema, output_dir: str | Path):
        """
        compile the schema to output directory

        :param schema:      the model schema
        :param output_dir:  output directory where to put generated files
        :return:            the outputs (self.outputs)
        """
        self.set_schema(schema)
        self.filtered_output_classes = {o.TYPE: o for o in self._filter_classes(self.output_classes, 'output')}
        self.output_dir = Path(output_dir).absolute()  # noqa
        self.outputs = {}
        self.output_builders = [e(self) for e in self._filter_classes(self.output_builder_classes, 'builder')]
        self.property_preprocessors = [e(self) for e in
                                       self._filter_classes(self.property_preprocessor_classes, 'property')]

        for model_preprocessor in self._filter_classes(self.model_preprocessor_classes, 'model'):
            model_preprocessor(self).transform(schema, schema.settings)

        # process the file
        self._iterate_schema(schema)

        for output in sorted(self.outputs.values(), key=lambda x: x.path):
            output.finish()

        return self.outputs

    def set_schema(self, schema):
        self.schema = schema
        self.settings = schema.settings

    # private methods

    def _filter_classes(self, classes: List[Type[object]], plugin_type):
        plugin_config = self.settings.plugins.get(plugin_type, None)
        if not plugin_config:
            return classes
        disabled = plugin_config.get('disabled', [])
        enabled = plugin_config.get('enabled', [])

        if disabled == '__all__':
            ret = []
        elif isinstance(disabled, (list, tuple)):
            ret = [c for c in classes if c.TYPE not in disabled]
        else:
            raise AttributeError('Value of settings.plugin.*.disabled must be either '
                                 'a list of names or string __all__ to disable all plugins.')
        if enabled:
            ret.extend(
                [c for c in classes if c.TYPE in enabled]
            )
            # add directly imported classes
            for en in enabled:
                if ':' in en:
                    en = en.split(':', maxsplit=1)
                    ret.append(getattr(importlib.import_module(en[0]), en[1]))
        return ret

    def _iterate_schema(self, schema: ModelSchema):
        for output_builder in self.output_builders:
            output_builder.begin(schema, schema.settings)

            for proc in self.property_preprocessors:
                proc.begin(schema, schema.settings)

            self._iterate_schema_output_builder(schema, output_builder)

            for proc in self.property_preprocessors:
                proc.finish()

            output_builder.finish()

    def _iterate_schema_output_builder(self, schema: ModelSchema, output_builder: OutputBuilder):
        def call_processors(stack, output_builder):
            data = copy.deepcopy(stack.top.data)
            for property_preprocessor in self.property_preprocessors:
                data = property_preprocessor.process(output_builder.TYPE, data, stack) or data
            return data

        def on_element(stack):
            data = call_processors(stack, output_builder)
            if isinstance(data, ReplaceElement):
                return data
            stack.top.data = data
            return output_builder.process_element(stack)

        ModelBuilderStack(schema).process(on_element)



try:
    import json5
except ImportError:
    import json as json5


def json_loader(file_path, schema):
    with open(file_path) as f:
        return json5.load(f)


json_loader.safe = True


def yaml_loader(file_path, schema):
    try:
        import yaml
    except ImportError:
        raise Exception('Loader for yaml not found. Please install pyyaml library to use yaml files')

    with open(file_path) as f:
        return yaml.full_load(f)


yaml_loader.safe = True


def python_loader(file_path, schema):
    with open(file_path) as f:
        code = f.read()
    # hope that user knows what he/she is doing
    gls = {}
    exec(code, gls)
    ret = {}
    for k, v in gls:
        try:
            json5.dumps(v)
        except:
            continue
        ret[k] = v
    return ret



import pkg_resources

from oarepo_model_builder.builder import ModelBuilder


def create_builder_from_entrypoints():
    output_classes = load_entry_points_list('oarepo_model_builder.ouptuts')
    builder_classes = load_entry_points_list('oarepo_model_builder.builders')
    preprocess_classes = load_entry_points_list('oarepo_model_builder.property_preprocessors')
    model_preprocessor_classes = load_entry_points_list('oarepo_model_builder.model_preprocessors')

    return ModelBuilder(
        output_builders=builder_classes,
        outputs=output_classes,
        property_preprocessors=preprocess_classes,
        model_preprocessors=model_preprocessor_classes
    )


def load_entry_points_dict(name):
    return {ep.name: ep.load() for ep in pkg_resources.iter_entry_points(group=name)}


def load_entry_points_list(name):
    ret = [(ep.name, ep.load()) for ep in pkg_resources.iter_entry_points(group=name)]
    ret.sort()
    return [x[1] for x in ret]



from oarepo_model_builder.utils.deepmerge import deepmerge
from oarepo_model_builder.model_preprocessors import ModelPreprocessor
from oarepo_model_builder.utils.camelcase import camel_case, snake_case


def last_item(x):
    return x.rsplit('.', maxsplit=1)[-1]


class InvenioModelPreprocessor(ModelPreprocessor):
    TYPE = 'invenio'

    def transform(self, schema, settings):
        deepmerge(settings, {
            'python': {
                'record_prefix': camel_case(settings.package.rsplit('.', maxsplit=1)[-1]),
                # just make sure that the templates is always there
                'templates': {
                },
                'marshmallow': {
                    'mapping': {
                    }
                }
            }
        })

        settings.setdefault('top-level-metadata', True)

        record_prefix = settings.python.record_prefix
        self.set(settings.python, 'record-prefix-snake',
                 lambda: snake_case(settings.python.record_prefix))

        self.set(settings.python, 'record-class',
                 lambda: (f'{settings.package}.record.{record_prefix}Record'))
        self.set(settings.python, 'record-schema-class',
                 lambda: (f'{settings.package}.schema.{record_prefix}Schema'))
        self.set(settings.python, 'record-schema-metadata-class',
                 lambda: (f'{settings.package}.schema.{record_prefix}MetadataSchema'))
        self.set(settings.python, 'record-schema-metadata-alembic',
                 lambda: (f'{settings.package_base}'))
        self.set(settings.python, 'record-schema-metadata-poetry',
                 lambda: (f'{settings.package_base}'))
        self.set(settings.python, 'record-metadata-class',
                 lambda: (f'{settings.package}.metadata.{record_prefix}Metadata'))

        self.set(settings.python, 'record-mapping-poetry',
                 lambda: (f'{settings.package_base}'))

        self.set(settings.python, 'record-jsonschemas-poetry',
                 lambda: (f'{settings.package_base}'))

        self.set(settings.python, 'record-permissions-class',
                 lambda: (f'{settings.package}.permissions.{record_prefix}PermissionPolicy'))
        self.set(settings.python, 'record-dumper-class',
                 lambda: (f'{settings.package}.dumper.{record_prefix}Dumper'))
        self.set(settings.python, 'record-metadata-table-name',
                 lambda: f'{record_prefix.lower()}_metadata')
        self.set(settings.python, 'record-search-options-class',
                 lambda: (f'{settings.package}.search_options.{record_prefix}SearchOptions'))
        self.set(settings.python, 'record-service-config-class',
                 lambda: (f'{settings.package}.service_config.{record_prefix}ServiceConfig'))
        self.set(settings.python, 'record-service-class',
                 lambda: (f'{settings.package}.service.{record_prefix}Service'))
        self.set(settings.python, 'record-resource-config-class',
                 lambda: (f'{settings.package}.resource.{record_prefix}ResourceConfig'))
        self.set(settings.python, 'record-resource-class',
                 lambda: (f'{settings.package}.resource.{record_prefix}Resource'))
        self.set(settings.python, 'record-resource-blueprint-name',
                 lambda: record_prefix)
        self.set(settings.python, 'register-blueprint-function',
                 lambda: (f'{settings.package}.blueprint.register_blueprint'))



from pathlib import Path

from oarepo_model_builder.builders import OutputBuilder
from oarepo_model_builder.builders.utils import ensure_parent_modules


class PythonBuilder(OutputBuilder):
    def module_to_path(self, module_name):
        mod = module_name.split('.')
        mod[-1] += '.py'
        return Path(*mod)

    def create_parent_modules(self, python_path):
        ensure_parent_modules(
            self.builder,
            python_path,
            max_depth=len(python_path.parts)
        )

    def class_to_path(self, class_name):
        return self.module_to_path(class_name.rsplit('.', maxsplit=1)[0])



from oarepo_model_builder.stack import ModelBuilderStack


class SchemaPathValidator:
    valid = False

    def get(self, path_el):
        raise NotImplementedError()


class Invalid(SchemaPathValidator):
    def get(self, path_el):
        return self


invalid = Invalid()


class PrimitiveValidator(SchemaPathValidator):
    valid = True

    def get(self, path_el):
        return invalid


primitive = PrimitiveValidator()


class DictValidator(SchemaPathValidator):
    valid = True

    def __init__(self, dict=None, primitives=None):
        self.dict = dict or {}
        if primitives:
            for p in primitives.split(','):
                p = p.strip()
                if not p:
                    continue
                self.dict[p] = primitive

    def get(self, path_el):
        if path_el in self.dict:
            return self.dict[path_el]
        return invalid


class AnyKeyDictValidator(SchemaPathValidator):
    valid = True

    def __init__(self, _next):
        self._next = _next

    def get(self, path_el):
        return self._next


class Ref(SchemaPathValidator):
    refs = {}
    valid = True

    def __init__(self, refstr):
        self.refstr = refstr

    def get(self, path_el):
        return self.refs[self.refstr].get(path_el)


Ref.refs['type'] = DictValidator({
    "properties": Ref('property'),
    "patternProperties": Ref('property'),
    "additionalProperties": DictValidator(primitives='type'),
    "propertyNames": DictValidator(primitives='pattern'),
    "items": Ref('type'),
    "prefixItems": Ref('type'),
    "contains": Ref('type'),
},
    primitives="type,"
               "enum,const,"  # enums and consts
               "required,minProperties,maxProperties,"  # object
               "minItems,maxItems,uniqueItems,minContains,maxContains,"  # array
               "minLength,maxLength,pattern,format,"  # string
               "minimum,exclusiveMinimum,maximum,exclusiveMaximum,multipleOf"  # numbers
)

Ref.refs['property'] = AnyKeyDictValidator(Ref('type'))

schema_paths = DictValidator(
    {
        "$vocabulary": AnyKeyDictValidator(primitive),
        "properties": Ref('property')
    }, primitives='$schema,$id,type'
)

model_paths = DictValidator(
    {
        'model': schema_paths
    }
)


def is_schema_element(stack: ModelBuilderStack):
    sc = model_paths
    for entry in stack.stack:
        key = entry.key
        if key is None:
            continue
        if isinstance(key, int):
            continue
        sc = sc.get(key)
        if not sc.valid:
            return False
    return True



from oarepo_model_builder.builders.python import PythonBuilder
from oarepo_model_builder.outputs.python import PythonOutput
from oarepo_model_builder.utils.hyphen_munch import HyphenMunch


class InvenioBaseClassPythonBuilder(PythonBuilder):
    class_config = None
    template = None

    def finish(self, **extra_kwargs):
        python_path = self.class_to_path(self.settings.python[self.class_config])
        self.create_parent_modules(python_path)

        output: PythonOutput = self.builder.get_output(
            'python',
            python_path
        )

        output.merge(
            self.template,
            HyphenMunch(settings=self.settings, python=self.settings.python, **extra_kwargs)
        )



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['oarepo_model_builder',
 'oarepo_model_builder.builders',
 'oarepo_model_builder.invenio',
 'oarepo_model_builder.loaders',
 'oarepo_model_builder.model_preprocessors',
 'oarepo_model_builder.outputs',
 'oarepo_model_builder.property_preprocessors',
 'oarepo_model_builder.templates',
 'oarepo_model_builder.utils']

package_data = \
{'': ['*'], 'oarepo_model_builder.invenio': ['templates/*']}

install_requires = \
['Jinja2>=3.0.3,<4.0.0',
 'black>=21.11b1,<22.0',
 'click>=7.1',
 'deepdiff>=5.6.0,<6.0.0',
 'isort>=5.10.1,<6.0.0',
 'jsonpointer>=2.2,<3.0',
 'libcst>=0.3.19,<0.4.0',
 'munch>=2.5.0,<3.0.0',
 'tomlkit>=0.7.2,<0.8.0']

extras_require = \
{'json5': ['json5>=0.9.6,<0.10.0'], 'pyyaml': ['PyYAML>=6.0,<7.0']}

entry_points = \
{'console_scripts': ['oarepo-compile-model = oarepo_model_builder.cli:run'],
 'oarepo_model_builder.builders': ['020-jsonschema = '
                                   'oarepo_model_builder.builders.jsonschema:JSONSchemaBuilder',
                                   '030-mapping = '
                                   'oarepo_model_builder.builders.mapping:MappingBuilder',
                                   '100-python_structure = '
                                   'oarepo_model_builder.builders.python_structure:PythonStructureBuilder',
                                   '110-invenio_record = '
                                   'oarepo_model_builder.invenio.invenio_record:InvenioRecordBuilder',
                                   '120-invenio_record_metadata = '
                                   'oarepo_model_builder.invenio.invenio_record_metadata:InvenioRecordMetadataBuilder',
                                   '130-invenio_record_schema = '
                                   'oarepo_model_builder.invenio.invenio_record_schema:InvenioRecordSchemaBuilder',
                                   '200-invenio_record_permissions = '
                                   'oarepo_model_builder.invenio.invenio_record_permissions:InvenioRecordPermissionsBuilder',
                                   '300-invenio_record_search_options = '
                                   'oarepo_model_builder.invenio.invenio_record_search:InvenioRecordSearchOptionsBuilder',
                                   '310-invenio_record_service_config = '
                                   'oarepo_model_builder.invenio.invenio_record_service_config:InvenioRecordServiceConfigBuilder',
                                   '320-invenio_record_service = '
                                   'oarepo_model_builder.invenio.invenio_record_service:InvenioRecordServiceBuilder',
                                   '340-invenio_record_dumper = '
                                   'oarepo_model_builder.invenio.invenio_record_dumper:InvenioRecordDumperBuilder',
                                   '400-invenio_record_resource_config = '
                                   'oarepo_model_builder.invenio.invenio_record_resource_config:InvenioRecordResourceConfigBuilder',
                                   '410-invenio_record_resource = '
                                   'oarepo_model_builder.invenio.invenio_record_resource:InvenioRecordResourceBuilder',
                                   '420-invenio_blueprint = '
                                   'oarepo_model_builder.invenio.invenio_blueprint:InvenioBlueprintBuilder',
                                   '900-invenio_sample_app_poetry = '
                                   'oarepo_model_builder.invenio.invenio_sample_app_poetry:InvenioSampleAppPoetryBuilder',
                                   '910-invenio_record_metadata_alembic_poetry '
                                   '= '
                                   'oarepo_model_builder.invenio.invenio_record_metadata_alembic_poetry:InvenioRecordMetadataAlembicPoetryBuilder',
                                   '920-invenio_record_metadata_models_poetry '
                                   '= '
                                   'oarepo_model_builder.invenio.invenio_record_metadata_models_poetry:InvenioRecordMetadataModelsPoetryBuilder',
                                   '930-invenio_resource_poetry = '
                                   'oarepo_model_builder.invenio.invenio_record_resource_poetry:InvenioRecordResourcePoetryBuilder',
                                   '940-invenio_record_search_poetry = '
                                   'oarepo_model_builder.invenio.invenio_record_search_poetry:InvenioRecordSearchPoetryBuilder',
                                   '950-invenio_record_jsonschemas_poetry = '
                                   'oarepo_model_builder.invenio.invenio_record_jsonschemas_poetry:InvenioRecordJSONSchemasPoetryBuilder'],
 'oarepo_model_builder.loaders': ['json = '
                                  'oarepo_model_builder.loaders:json_loader',
                                  'json5 = '
                                  'oarepo_model_builder.loaders:json_loader',
                                  'yaml = '
                                  'oarepo_model_builder.loaders:yaml_loader',
                                  'yml = '
                                  'oarepo_model_builder.loaders:yaml_loader'],
 'oarepo_model_builder.model_preprocessors': ['01-default = '
                                              'oarepo_model_builder.model_preprocessors.default_values:DefaultValuesModelPreprocessor',
                                              '10-invenio = '
                                              'oarepo_model_builder.model_preprocessors.invenio:InvenioModelPreprocessor',
                                              '20-elasticsearch = '
                                              'oarepo_model_builder.model_preprocessors.elasticsearch:ElasticsearchModelPreprocessor'],
 'oarepo_model_builder.ouptuts': ['jsonschema = '
                                  'oarepo_model_builder.outputs.jsonschema:JSONSchemaOutput',
                                  'mapping = '
                                  'oarepo_model_builder.outputs.mapping:MappingOutput',
                                  'python = '
                                  'oarepo_model_builder.outputs.python:PythonOutput',
                                  'toml = '
                                  'oarepo_model_builder.outputs.toml:TOMLOutput'],
 'oarepo_model_builder.property_preprocessors': ['10-text_keyword = '
                                                 'oarepo_model_builder.property_preprocessors.text_keyword:TextKeywordPreprocessor'],
 'oarepo_model_builder.templates': ['99-base_templates = '
                                    'oarepo_model_builder.invenio']}

setup_kwargs = {
    'name': 'oarepo-model-builder',
    'version': '0.9.0',
    'description': 'An utility library that generates OARepo required data model files from a JSON specification file',
    'long_description': '# OARepo model builder\n\nA library and command-line tool to generate invenio model from a single model file.\n\n## CLI Usage\n\n```bash\noarepo-compile-model model.yaml\n```\n\nwill compile the model.yaml into the current directory. Options:\n\n```bash\n  --output-directory <dir> Output directory where the generated files will be\n                           placed. Defaults to "."\n  --package <name>         Package into which the model is generated. If not\n                           passed, the name of the current directory,\n                           converted into python package name, is used.\n  --set <name=value>       Overwrite option in the model file. \n                           Example --set settings.elasticsearch.keyword-ignore-above=20\n  -v                       Increase the verbosity. This option can be used\n                           multiple times.\n  --config <filename>      Load a config file and replace parts of the model\n                           with it. The config file can be a json, yaml or a\n                           python file. If it is a python file, it is\n                           evaluated with the current model stored in the\n                           "oarepo_model" global variable and after the\n                           evaluation all globals are set on the model.\n  --isort / --skip-isort   Call isort on generated sources (default: yes)\n  --black / --skip-black   Call black on generated sources (default: yes)\n```\n\n## Model file structure\n\nA model is a json/yaml file with the following structure:\n\n```yaml\nsettings:\n  python:\n  elasticsearch:\nmodel:\n  properties:\n    title: { type: \'fulltext\' }\n```\n\nThere might be more sections (documentation etc.), but only the ``settings`` and ``model`` are currently processed.\n\n### settings section\n\nThe settings section might contain the following keys\n(default values below):\n\n```yaml\nsettings:\n  package: basename(output dir) with \'-\' converted to \'_\'\n  kebap-package: to_kebap(package)\n  package-path: path to package as python Path instance\n  schema-version: 1.0.0\n  schema-name: { kebap-package }-{schema-version}.json\n  schema-file: full path to generated json schema\n  mapping-file: full path to generated mapping\n  collection-url: camel_case(last component of package)\n\n  processing-order: [ \'settings\', \'*\', \'model\' ]\n\n  python:\n    record-prefix: camel_case(last component of package)\n    templates: { }   # overridden templates\n    marshmallow:\n      top-level-metadata: true\n      mapping: { }\n\n    record-prefix-snake: snake_case(record_prefix)\n\n    record-class: { settings.package }.record.{record_prefix}Record\n      # full record class name with package\n    record-schema-class: { settings.package }.schema.{record_prefix}Schema\n      # full record schema class name (apart from invenio stuff, contains only metadata field)\n    record-schema-metadata-class: { settings.package }.schema.{record_prefix}MetadataSchema\n      # full record schema metadata class name (contains model schema as marshmallow)\n    record-schema-metadata-alembic: { settings.package_base }\n    # name of key in pyproject.toml invenio_db.alembic entry point \n    record-metadata-class: { settings.package }.metadata.{record_prefix}Metadata\n      # db class to store record\'s metadata \n    record-metadata-table-name: { record_prefix.lower() }_metadata\n      # name of database table for storing metadata \n    record-permissions-class: { settings.package }.permissions.{record_prefix}PermissionPolicy\n      # class containing permissions for the record\n    record-dumper-class: { settings.package }.dumper.{record_prefix}Dumper\n      # record dumper class for elasticsearch\n    record-search-options-class: { settings.package }.search_options.{record_prefix}SearchOptions\n      # search options for the record\n    record-service-config-class: { settings.package }.service_config.{record_prefix}ServiceConfig\n      # configuration of record\'s service\n    record-resource-config-class: { settings.package }.resource.{record_prefix}ResourceConfig\n      # configuration of record\'s resource\n    record-resource-class: { settings.package }.resource.{record_prefix}Resource\n      # record resource\n    record-resource-blueprint-name: { record_prefix }\n    # blueprint name of the resource \n    register-blueprint-function: { settings.package }.blueprint.register_blueprint\'\n      # name of the blueprint registration function\n\n  elasticsearch:\n    keyword-ignore-above: 50\n\n  plugins:\n    packages: [ ]\n    # list of extra packages that should be installed in compiler\'s venv\n    output|builder|model|property:\n      # plugin types - file outputs, builders, model preprocessors, property preprocessors \n      disabled: [ ]\n      # list of plugin names to disable\n      # string "__all__" to disable all plugins in this category    \n      enabled:\n      # list of plugin names to enable. The plugins will be used\n      # in the order defined. Use with disabled: __all__\n      # list of "module:className" that will be added at the end of\n      # plugin list\n```\n\n### model section\n\nThe model section is a json schema that might be annotated with extra sections. For example:\n\n```yaml\nmodel:\n  properties:\n    title:\n      type: multilingual\n      oarepo:ui:\n        label: Title\n        class: bold-text\n      oarepo:documentation: |\n        Lorem ipsum ...\n        Dolor sit ...\n```\n\n**Note**: ``multilingual`` is a special type (not defined in this library) that is translated to the correct schema,\nmapping and marshmallow files with a custom ``PropertyPreprocessor``.\n\n``oarepo:ui`` gives information for the ui output\n\n``oarepo:documentation`` is a section that is currently ignored\n\n## Referencing a model\n\n## API Usage\n\nTo generate invenio model from a model file, perform the following steps:\n\n1. Load the model into a ``ModelSchema`` instance\n    ```python\n    from oarepo_model_builder.schema import ModelSchema\n    from oarepo_model_builder.loaders import yaml_loader\n   \n    included_models = {\n        \'my_model\': lambda parent_model: {\'test\': \'abc\'} \n    }\n    loaders = {\'yaml\': yaml_loader}\n   \n    model = ModelSchema(file_path=\'test.yaml\', \n                        included_models=included_models, \n                        loaders=loaders)\n    ```\n\n   You can also path directly the content of the file path in ``content`` attribute\n\n   The ``included_models`` is a mapping between model key and its accessor. It is used to replace any ``oarepo:use``\n   element. See the Referencing a model above.\n\n   The ``loaders`` handle loading of files - the key is lowercased file extension, value a function taking (schema,\n   path) and returning loaded content\n\n\n2. Create an instance of ``ModelBuilder``\n\n   To use the pre-installed set of builders and preprocessors, invoke:\n\n   ```python\n   from oarepo_model_builder.entrypoints \\ \n    import create_builder_from_entrypoints\n   \n   builder = create_builder_from_entrypoints()\n   ```\n\n   To have a complete control of builders and preprocessors, invoke:\n\n   ```python\n      from oarepo_model_builder.builder import ModelBuilder\n      from oarepo_model_builder.builders.jsonschema import JSONSchemaBuilder\n      from oarepo_model_builder.builders.mapping import MappingBuilder\n      from oarepo_model_builder.outputs.jsonschema import JSONSchemaOutput\n      from oarepo_model_builder.outputs.mapping import MappingOutput\n      from oarepo_model_builder.outputs.python import PythonOutput\n      from oarepo_model_builder.property_preprocessors.text_keyword import TextKeywordPreprocessor\n      from oarepo_model_builder.model_preprocessors.default_values import DefaultValuesModelPreprocessor\n      from oarepo_model_builder.model_preprocessors.elasticsearch import ElasticsearchModelPreprocessor\n\n      builder = ModelBuilder(\n        output_builders=[JSONSchemaBuilder, MappingBuilder],\n        outputs=[JSONSchemaOutput, MappingOutput, PythonOutput],\n        model_preprocessors=[DefaultValuesModelPreprocessor, ElasticsearchModelPreprocessor],\n        property_preprocessors=[TextKeywordPreprocessor]\n      )    \n   ```   \n\n\n3. Invoke\n\n   ```python\n      builder.build(schema, output_directory)\n   ```\n\n## Extending the builder\n\n### Builder pipeline\n\n![Pipeline](./docs/oarepo-model-builder.png)\n\nAt first, an instance of [ModelSchema](./oarepo_model_builder/schema.py) is obtained. The schema can be either passed\nthe content of the schema as text, or just a path pointing to the file. The extension of the file determines\nwhich [loader](./oarepo_model_builder/loaders/__init__.py) is used. JSON, JSON5 and YAML are supported out of the box (\nif you have json5 and pyyaml packages installed)\n\nThen [ModelBuilder](./oarepo_model_builder/builder.py).build(schema, output_dir) is called.\n\nIt begins with calling all [ModelPreprocessors](./oarepo_model_builder/model_preprocessors/__init__.py). They get the\nwhole schema and settings and can modify both.\nSee [ElasticsearchModelPreprocessor](./oarepo_model_builder/model_preprocessors/elasticsearch.py) as an example. The\ndeepmerge function does not overwrite values if they already exist in settings.\n\nFor each of the outputs (jsonschema, mapping, record, resource, ...)\nthe top-level properties of the transformed schema are then iterated. \nThe order of the top-level properties is given by ``settings.processing-order``.\n\nThe top-level property and all its descendants (a visitor patern, visiting property by property), \na [PropertyPreprocessor](./oarepo_model_builder/property_preprocessors/__init__.py)\nis called. \n\nThe preprocessor can either modify the property, decide to remove it or replace it with a new set of properties\n(see [multilang in tests](./tests/multilang.py) ).\n\nThe property is then passed to the\n[OutputBuilder](./oarepo_model_builder/builders/__init__.py)\n(an example is [JSONSchemaBuilder](./oarepo_model_builder/builders/jsonschema.py))\nthat serializes the tree of properties into the output.\n\nThe output builder does not create files on the filesystem explicitly but uses instances\nof [OutputBase](./oarepo_model_builder/outputs/__init__.py), for\nexample [JSONOutput](./oarepo_model_builder/outputs/json.py) or more\nspecialized [JSONSchemaOutput](./oarepo_model_builder/outputs/jsonschema.py).\n\nSee [JSONBaseBuilder](./oarepo_model_builder/builders/json_base.py) for an example of how to get an output and write to\nit (in this case, the json-based output).\n\nThis way, even if more output builders access the same file, their access is coordinated.\n\n### Registering Preprocessors, Builders and Outputs for commandline client\n\nThe model & property preprocessors, output builders and outputs are registered in entry points. In poetry, it looks as:\n\n```toml\n[tool.poetry.plugins."oarepo_model_builder.builders"]\n010-jsonschema = "oarepo_model_builder.builders.jsonschema:JSONSchemaBuilder"\n020-mapping = "oarepo_model_builder.builders.mapping:MappingBuilder"\n030-python_structure = "oarepo_model_builder.builders.python_structure:PythonStructureBuilder"\n040-invenio_record = "oarepo_model_builder.invenio.invenio_record:InvenioRecordBuilder"\n\n[tool.poetry.plugins."oarepo_model_builder.ouptuts"]\njsonschema = "oarepo_model_builder.outputs.jsonschema:JSONSchemaOutput"\nmapping = "oarepo_model_builder.outputs.mapping:MappingOutput"\npython = "oarepo_model_builder.outputs.python:PythonOutput"\n\n[tool.poetry.plugins."oarepo_model_builder.property_preprocessors"]\n010-text_keyword = "oarepo_model_builder.preprocessors.text_keyword:TextKeywordPreprocessor"\n\n[tool.poetry.plugins."oarepo_model_builder.model_preprocessors"]\n01-default = "oarepo_model_builder.transformers.default_values:DefaultValuesModelPreprocessor"\n10-invenio = "oarepo_model_builder.transformers.invenio:InvenioModelPreprocessor"\n20-elasticsearch = "oarepo_model_builder.transformers.elasticsearch:ElasticsearchModelPreprocessor"\n\n[tool.poetry.plugins."oarepo_model_builder.loaders"]\njson = "oarepo_model_builder.loaders:json_loader"\njson5 = "oarepo_model_builder.loaders:json_loader"\nyaml = "oarepo_model_builder.loaders:yaml_loader"\nyml = "oarepo_model_builder.loaders:yaml_loader"\n\n[tool.poetry.plugins."oarepo_model_builder.templates"]\n99-base_templates = "oarepo_model_builder.invenio.templates"\n```\n\n### Generating python files\n\nThe default python output is based on [libCST](https://github.com/Instagram/LibCST) that enables merging generated code\nwith a code that is already present in output files. The transformer provided in this package can:\n\n1. Add imports\n2. Add a new class or function on top-level\n3. Add a new method to an existing class\n4. Add a new const/property to an existing class\n\nThe transformer will not touch an existing function/method. Increase verbosity level to get a list of rejected patches\nor add ``--set settings.python.overwrite=true``\n(use with caution, with sources stored in git and do diff afterwards).\n\n#### Overriding default templates\n\nThe default templates are written as jinja2-based templates.\n\nTo override a single or multiple templates, create a package containing the templates and register it\nin ``oarepo_model_builder.templates``. Be sure to specify the registration key smaller than ``99-``. The template loader\niterates the sorted set of keys and your templates would be loaded before the default ones. Example:\n\n   ```\n   my_package\n      +-- __init__.py\n      +-- templates\n          +-- invenio_record.py.jinja2 \n   ```\n\n   ```python\n   # my_package/__init__.py\nTEMPLATES = {\n    # resolved relative to the package\n    "record": "templates/invenio_record.py.jinja2"\n}\n   ```\n\n   ```toml\n   [tool.poetry.plugins."oarepo_model_builder.templates"]\n20-my_templates = "my_package"\n   ```\n\nTo override a template for a single model, in your model file (or configuration file with -c option or via --set option)\n, specify the relative path to the template:\n\n```yaml\nsettings:\n  python:\n    templates:\n      record: ./test/my_invenio_record.py.jinja2\n```\n',
    'author': 'Miroslav Bauer',
    'author_email': 'bauer@cesnet.cz',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'entry_points': entry_points,
    'python_requires': '>=3.10,<4.0',
}


setup(**setup_kwargs)
